"""Helper functions to scale values linearly or logarithmically.

Functions
---------
scale_value
linear
ln
log_10
log_2

Notes
-----
The functions in this module attempt to preserve the input type when possible.
The user will expect to see an integer when an integer is input, a float when
a float is input, or a string when a string is input. The only function where
integer preservation is not possible is `ln`.
"""
import numbers
import warnings

import numpy as np


def scale_value(value, scaling):
    """Scale a value.

    Parameters
    ----------
    value
        The value to scale.
    scaling : {'linear', 'ln', 'log_10', 'log_2'} or callable
        The name of the `shadho.scaling` function to use or a function/callable
        object to apply to value.

    Returns
    -------
    scaled
        `value` scaled by `scaling` or `value` if `scaling` cannot be applied.

    Notes
    -----
    In the case that `scaling` cannot be used (i.e. it is not the name of a
    valid `shadho.scaling` function and cannot be called), a warning is issued
    and the value is returned without modification.
    """
    SCALES = {
        'linear': linear,
        'ln': ln,
        'log_10': log_10,
        'log_2': log_2
    }
    try:
        scaled = SCALES[scaling](value) if scaling in SCALES \
                 else scaling(value)
    except TypeError:
        msg = "Invalid scaling {}. {} will not be altered.".format(scaling,
                                                                   value)
        warnings.warn(msg)
        scaled = value

    return scaled


def linear(x, coeff=1.0, degree=1.0):
    """Scale a value linearly.

    The passed value will be scaled according to
    :math:`coeff\,*\,x^{degree}`.

    Parameters
    ----------
    x
        The value to scale linearly.
    coeff : float
        Multiply `x` by this.
    degree : float
        Raise `x` to this power.

    Returns
    -------
    The scaled value, or `x` if non-numeric.

    Notes
    -----
    This function preserves the type of `x` using an explicit cast. Passing
    an int will return an int, a float will return a float, a string will
    return a string, etc.
    """
    try:
        if not isinstance(x, np.ndarray):
            t = type(x)
            x = t(coeff * np.power(x, degree))
        else:
            dtype = x.dtype
            x = (coeff * np.power(x, degree)).astype(dtype)
    except TypeError:
        x = x

    return x


def ln(x):
    """Scale exponentially to base *e*.

    Parameters
    ----------
    x
        The value to scale.

    Returns
    -------
    :math:`e^x`

    Raises
    ------
    TypeError
        Raised when `x` is non-numeric.

    Notes
    -----
    This function is the only scaling function that does not preserve
    the input type. All return values are converted to floating point
    as a consequence of using base *e*.
    """
    try:
        if not isinstance(x, numbers.Number):
            raise TypeError
        x = np.exp(x)
    except TypeError:
        x = x
    return x


def log_10(x):
    """Scale exponentially to base 10.

    This function attempts to maintain the type of the input value x. The
    typical use case for this type of scaling is to generate a range of integer
    powers of 10 (e.g, [1, 10, 100, 1000, 10000]), however floating-point
    values in range (0, 1) are also common.

    Parameters
    ----------
    x : numeric
        The value to scale.

    Returns
    -------
    :math:`10^x`

    Notes
    -----
    This function preserves the type of `x` using an explicit cast. Passing
    an int will return an int, a float will return a float, a string will
    return a string, etc.

    The exception to this rule is when an `x` less than 0 is supplied. In this
    case the return value will always be floating-point.
    """
    try:
        if isinstance(x, numbers.Number):
            t = type(x)
            x = np.power(10.0, x)
            if t(x) == x:
                x = t(x)
        elif isinstance(x, np.ndarray):
            dtype = x.dtype
            x = np.power(10.0, x).astype(dtype)
        else:
            raise TypeError('{} is not a numeric type, cannot take log'.format(x))
    except TypeError:
        x = x

    return x


def log_2(x):
    """Scale exponentially to base 2.

    This function attempts to maintain the type of the input value x. The
    typical use case for this type of scaling is to generate a range of integer
    powers of 2 (e.g, [16, 32, 64, 128, 256]).

    Parameters
    ----------
    x : numeric
        The value to scale.

    Returns
    -------
    :math:`2^x`

    Notes
    -----
    This function preserves the type of `x` using an explicit cast. Passing
    an int will return an int, a float will return a float, a string will
    return a string, etc.

    The exception to this rule is when an `x` less than 0 is supplied. In this
    case the return value will always be floating-point.
    """
    try:
        if isinstance(x, numbers.Number):
            t = type(x)
            x = np.exp2(x)
            if t(x) == x:
                x = t(x)
        elif isinstance(x, np.ndarray):
            dtype = x.dtype
            x = np.exp2(x).astype(x.dtype)
        else:
            raise TypeError('{}  is not a numeric type, cannot take log'.format(x))
    except TypeError:
        x = x

    return x



"""Facilities for grouping distributed hardware and search spaces.

Classes
-------
ComputeClass
    Group hardware by a common property.
"""
import uuid
import sys

from pyrameter.optimizer import FMin


class ComputeClass(object):
    """Group hardware by a common property.

    When working with heterogeneous distributed hardware (e.g., multiple
    clusters, multiple cloud services), it is often useful to group machines by
    a common resource. ComputeClass allows SHADHO to target specific hardware
    for running specific tasks. Hardware may be grouped by:

        * CPU count
        * Available RAM
        * GPU model
        * GPU count
        * Other accelerators
        * and by arbitrary, user-defined values.

    Parameters
    ----------
    name : str
        User-level name for the compute class (e.g. "8-core", "1080ti", etc.).
    resource : str
        The name of the resource common to this group.
    value
        The value of the resource common to this group.
    max_queued_tasks : int
        The maximum number of tasks to queue. Recommended to be 1.5-2x the
        number of expected nodes with this resource.

    Attributes
    ----------
    id : str
        Internal id of this ComputeClass.
    name : str
        User-level name for the compute class (e.g. "8-core", "1080ti", etc.).
    resource : str
        The name of the resource common to this group.
    value
        The value of the resource common to this group.
    max_queued_tasks : int
        The maximum number of tasks to queue. Recommended to be 1.5-2x the
        number of expected nodes with this resource.
    current_tasks : int
        The current number of queued tasks.
    model_group : `pyrameter.ModelGroup`
        The (possibly ordered) list of models assigned to this ComputeClass.

    See Also
    --------
    `pyrameter.ModelGroup`
    """
    def __init__(self, name, resource, value, max_queued_tasks, optimizer):
        self.id = str(uuid.uuid4())
        self.name = name
        self.resource = resource
        self.value = value
        self.max_queued_tasks = max_queued_tasks
        self.current_tasks = 0
        self.optimizer = optimizer

    def __hash__(self):
        return hash((self.id, self.name, self.resource, self.value))

    def generate(self, ssid=None):
        """Generate a set of hyperparameters from a model in this group.

        Hyperparameters are generated from a single model assigned to this
        compute class. If no model id is provided, the model is selected based
        on a probability distribution over the models.

        Parameters
        ----------
        ssid : optional
            If a valid id is supplied, generate hyperparameters values from the
            requested search space. Otherwise, probabilistically select a
            search space and generate hyperparameter values.
        """
        return self.optimizer.generate(ssid=ssid)

    def add_searchspace(self, searchspace):
        """Add a search space to this compute class.

        Parameters
        ----------
        searchspace : `pyrameter.searchspace.SearchSpace`
        """
        self.optimizer.searchspaces.append(searchspace)

    def remove_searchspace(self, ssid):
        """Remove a search space from this compute class.

        Parameters
        ----------
        model_id : str
        """
        idx = None
        for i, ss in enumerate(self.optimizer.searchspaces):
            if ss.id == ssid:
                idx = i
                break
        if idx is not None:
            self.optimizer.searchspaces.pop(idx)

    def clear(self):
        """Remove all models from this compute class."""
        self.optimizer.searchspaces = []

    def register_result(self, ssid, objective, result=None, errmsg=None):
        """Add a result to a model in this compute class.

        Parameters
        ----------
        model_id : str
            The id of the model to store the result in.

        """
        trial = self.optimizer.trials[ssid]
        trial.objective = objective
        trial.result = result
        trial.errmsg = errmsg



"""Configuration settings for `shadho`.

Classes
-------
ShadhoConfig
    Configurations for running SHADHO.
ConfigGroup
    Object-oriented interface for hierarchical config settings.

Notes
-----
`shadho` can be configured using the `.shadhorc` file, by passing arguments to
the `shadho.Shadho` object initializer, or by passing values to the `cfg`
attribute of an initialized `shadho.Shadho` object.
"""
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import copy
import os


class ShadhorcDoesNotExistError(Exception):
    def __init__(self, configfile):
        msg = "{} is not a valid shadho config file".format(configfile)
        super(ShadhorcDoesNotExistError, self).__init__(msg)


class ShadhoInstallNotfoundError(Exception):
    def __init__(self, directory):
        msg = "Could not find shadho installation at {}".format(directory)
        super(ShadhoInstallNotfoundError, self).__init__(msg)


class ShadhoConfig(object):
    """Configurations for running SHADHO.

    Configurations consist of the defaults defined in this class, those values
    read from the .shahdorc file, and those set by the user at runtime.

    Parameters
    ----------
    use_defaults : bool, optional
        If true, only use the default configuration. Default: False.

    Attributes
    ----------
    config : dict
        Dictionary of configuration values.

    Notes
    -----
    All config values can be addressed by key index (dictionary style) or by
    dot operator (object-oriented style).

    """

    DEFAULTS = {
        'global': {
            'wrapper': 'shadho_worker.py',
            'utils': 'shadho_utils.py',
            'output': 'out.tar.gz',
            'result_file': 'performance.json',
            'optimize': 'loss',
            'param_file': 'hyperparameters.json',
            'backend': 'json',
            'manager': 'workqueue',
        },
        'workqueue': {
            'port': str(9123),
            'name': 'shadho_master',
            'shutdown': str(True),
            'logfile': 'shadho_master.log',
            'debugfile': 'shadho_master.debug',
            'password': str(False)
        },
        'backend': {
            'type': 'sql',
            'url': 'sqlite:///:memory:'
        }
    }

    def __init__(self, use_defaults=False):

        # Copy the defaults
        if 'shadho_dir' not in ShadhoConfig.DEFAULTS['global']:
            shadho_dir = os.path.join(self.__get_home(), '.shadho')
            ShadhoConfig.DEFAULTS['global']['shadho_dir'] = shadho_dir

        self.config = copy.deepcopy(ShadhoConfig.DEFAULTS)

        if not use_defaults:
            # Get the path to the shadhorc file
            configfile = os.getenv('SHADHORC') if 'SHADHORC' in os.environ \
                         else os.path.join(self.__get_home(), '.shadhorc')
            if not os.path.isfile(configfile):
                raise ShadhorcDoesNotExistError(configfile)

            self.configfile = configfile

            # Load custom settings
            cfg = configparser.ConfigParser()
            with open(configfile, 'r') as f:
                try:
                    cfg.read_file(f)
                except AttributeError:
                    cfg.readfp(f)

            for section in cfg.sections():
                for option in cfg.options(section):
                    try:
                        t = type(ShadhoConfig.DEFAULTS[section][option])
                    except KeyError:
                        t = str

                    if t is bool:
                        val = cfg.getboolean(section, option)
                    elif t is int:
                        val = cfg.getint(section, option)
                    elif t is float:
                        val = cfg.getfloat(section, option)
                    else:
                        val = cfg.get(section, option)

                    self.config[section][option] = val

        if not os.path.isdir(self.config['global']['shadho_dir']):
            raise ShadhoInstallNotfoundError(self.config['global']['shadho_dir'])

        # Instantiate config group objects
        self.shadho = ConfigGroup(self.config['global'])
        self.workqueue = ConfigGroup(self.config['workqueue'])
        self.backend = ConfigGroup(self.config['backend'])

    def __getattr__(self, attr):
        if attr not in self.__dict__:
            if attr in self.__dict__['shadho']:
                return getattr(self.__dict__['shadho'], attr)
            else:
                msg = '{} has no attribute {}'
                raise AttributeError(msg.format(self.__class__.__name__, attr))
        else:
            return self.__dict__[attr]

    def __get_home(self):
        """Get the absolute path to the user's home directory.

        Returns
        -------
        home : str
            Absolute path to the user's home directory.

        Notes
        -----
        This method attempts to be OS-agnostic.
        """
        try:
            home = os.path.expanduser(
                    os.environ['HOME'] if 'HOME' in os.environ
                    else os.environ['USERPROFILE'])
        except KeyError:
            print('Error: Could not find home directory in environment')
            home = ''

        return home

    def save_config(self, path):
        """Write the current config to file.

        Parameters
        ----------
        path : str
            Path to the output config file.
        """
        try:
            config = configparser.ConfigParser()
            config.read_dict(self.config)
        except AttributeError:
            config = configparser.RawConfigParser()
            for section in self.config:
                config.add_section(str(section))
                for entry in self.config[section]:
                    if type(self.config[section]) is dict:
                        config.set(str(section),
                                   str(entry),
                                   str(self.config[section][entry]))
        with open(os.path.join(path, '.shadhorc'), 'w') as f:
            config.write(f)


class ConfigGroup(object):
    """Container class for hierarchical configuration.

    Parameters
    ----------
    data : dict
        Key/value pairs for configuration values.

    Notes
    -----
    Configuration values should be accglobalessed using the ``.`` operator.

    """

    def __init__(self, data=None, **kws):
        if data is None:
            data = {}
        data.update(kws)
        self.data = data

    def __getattr__(self, attr):
        if attr != 'data':
            if attr in self.__dict__['data']:
                return self.__dict__['data'][attr]
            else:
                msg = '{} has no attribute {}'
                raise AttributeError(msg.format(self.__class__.__name__, attr))
        else:
            return self__dict__['data']

    def __setattr__(self, attr, val):
        if attr != 'data':
            self.__dict__['data'][attr] = val
        else:
            self.__dict__[attr] = val

    def __contains__(self, key):
        return key in self.data

    def __str__(self):
        return str(self.data)



"""Routines for automatically running functions and commands through `shadho`.

Functions
---------
load_config
    Load configurations sent from the `shadho` master.
run_task
    Run the user-defined task and save the result to file.
"""
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import copy
import json
import os
try:
    import cPickle as pickle
except ImportError:
    import pickle
import tarfile

# If running on a remote system, try to import from the clobbered module name.
# Otherwise, import from the standard module name.
try:
    from shadho_utils import ShadhoEncoder, ShadhoDecoder
except ImportError:
    from utils import ShadhoEncoder, ShadhoDecoder


DEFAULTS = {
    'global': {
        'wrapper': 'shadho_run_task.py',
        'output': 'out.tar.gz',
        'result_file': 'performance.json',
        'optimize': 'loss',
        'param_file': 'hyperparameters.json'
    }
}


def load_config(path='.shadhorc'):
    """Load the configuration for this task.

    In most cases, the ``path`` parameter should be omitted in favor of using
    the configuration provided by the `shadho` master. This will ensure that
    the local configuration matches the global configuration and reduce
    headaches.

    Parameters
    ----------
    path : str, optional
        Path to the configuration file. It is recommended to use the default
        value in most cases. This will load the configuration provided by the
        `shadho` master and ensure that the local configuration matches the
        global configuration.

    Returns
    -------
    cfg : dict
        The configuration to use with this task.
    """
    config = copy.deepcopy(DEFAULTS)

    if os.path.isfile(path):
        cfg = configparser.ConfigParser()
        with open(path, 'r') as f:
            try:
                cfg.read_file(f)
            except AttributeError:
                cfg.readfp(f)
        for section in cfg.sections():
            for option in cfg.options(section):
                try:
                    t = type(DEFAULTS[section][option])
                except KeyError:
                    continue
                if t is bool:
                    val = cfg.getboolean(section, option)
                elif t is int:
                    val = cfg.getint(section, option)
                elif t is float:
                    val = cfg.getfloat(section, option)
                else:
                    val = cfg.get(section, option)

                config[section][option] = val

    return config


def run(task, cfgpath='.shadhorc'):
    """Run a function with arguments pulled from a JSON source.

    For simple use cases, SHADHO should be able to run without user-defined
    bash scripts, JSON parsing, etc. Users will expect this to be automated in
    the standard case.

    Parameters
    ----------
    task : callable
        The function/callable object to run.
    cfgpath : str, optional
        The path to the `shadho` configuration file for this task.
    """
    try:
        # Load the configuration
        cfg = load_config(path=cfgpath)

        # Load hyperparameters to test
        spec = {}
        if os.path.isfile(cfg['global']['param_file']):
            with open(cfg['global']['param_file'], 'r') as f:
                spec = json.load(f, cls=ShadhoDecoder)

        # Run the task and save the results in a format `shahdo` recognizes.
        result = task(spec)
        if isinstance(result, float):
            result = {cfg['global']['optimize']: result}

        # Dump the results to file.
        with open(cfg['global']['result_file'], 'w') as f:
            json.dump(result, f, cls=ShadhoEncoder)

        # Compress into the expected output file.
        ext = os.path.splitext(cfg['global']['output'])
        mode = ':'.join(['w', ext]) if ext in ['gz', 'bz2'] else 'w'
        with tarfile.open(cfg['global']['output'], mode) as f:
            f.add(cfg['global']['result_file'])

    except IOError as err:
        print(err)
    except ValueError as err:
        print("Error decoding parameters in {}"
              .format(cfg['global']['param_file']))
        print(err)
    except TypeError as err:
        print(err)



#!/usr/bin/env python3
"""Install script for Work Queue and SHADHO config.

This executable performs post-install steps to get the Work Queue library and
set up SHADHO config and extra scripts. By default, these are installed to
`$HOME/.shadhorc` and `$HOME/.shadho/`.
"""

import argparse
import configparser
import os
import pathlib
import shutil
import site
import subprocess
import sys
import tempfile


def parse_args(args=None):
    homedir = str(pathlib.Path().home())

    p = argparse.ArgumentParser(
        description="""
            Install the .shadho support directory, Work Queue, and additional
            files needed to run SHADHO.
        """)

    p.add_argument('--prefix', type=str, default=homedir,
                   help='install directory of the .shadho directory')
    p.add_argument('--user', action='store_true',
                   help='pass to perform a Python user-level install')

    return p.parse_args(args=None)


def install_workqueue(prefix, user=False):
    """Install Work Queue and SHADHO supplementary files.

    Parameters
    ----------
    prefix : str
        Install directory of the .shadho directory.

    Returns
    -------
    prefix : str or None
        Identical to ``prefix``. Set to None if intallation failed.
    site_packages_prefix : str
        Path to the site-packages directory where work_queue was installed.
    """
    try:
        if os.path.basename(prefix) != '.shadho':
            prefix = os.path.join(prefix, '.shadho')
        tempdir = tempfile.mkdtemp(prefix='shadho_install.')

        os.makedirs(prefix, exist_ok=True)
        print('Building and installing Work Queue to {}.'.format(prefix))

        installer = os.path.join(os.path.dirname(__file__),
                                 'install_workqueue.sh')
        subprocess.check_output(
            ['sh', installer, tempdir, prefix],
            stderr=subprocess.STDOUT)
        result = prefix

        maj = sys.version_info.major
        min = sys.version_info.minor

        sp_prefix = os.path.join('lib',
                                 'python{}.{}'.format(maj, min),
                                 'site-packages')

        if user:
            dest_prefix = os.path.join(site.USER_BASE, sp_prefix)
        else:
            dest_prefix = os.path.join(sys.prefix, sp_prefix)

        shutil.copy(os.path.join(prefix, sp_prefix, 'work_queue.py'),
                    dest_prefix)
        shutil.copy(os.path.join(prefix, sp_prefix, '_work_queue.so'),
                    dest_prefix)

    except subprocess.CalledProcessError as e:
        logfile = os.path.abspath('shadho_install.log')
        print('Error installing Work Queue.')
        print('Dumping install logs to {}'.format(logfile))
        with open(logfile, 'w') as f:
            f.write(e.output.decode())
        result = None
        dest_prefix = None
    finally:
        shutil.rmtree(tempdir)

    return result, dest_prefix


def write_shadhorc(prefix, shadho_dir):
    """Create the .shadhorc config file.

    Parameters
    ----------
    prefix : str
        Path to install SHADHO config and Work Queue to.
    shadho_dir : str
        Path to the .shadho install directory.
    """
    # Set up the default config values
    default_config = {
        'global': {
            'wrapper': 'shadho_worker.py',
            'utils': 'shadho_utils.py',
            'output': 'out.tar.gz',
            'result_file': 'performance.json',
            'optimize': 'loss',
            'param_file': 'hyperparameters.json',
            'backend': 'json',
            'manager': 'workqueue'
        },
        'workqueue': {
            'port': '9123',
            'name': 'shadho_master',
            'shutdown': 'True',
            'logfile': 'shadho_master.log',
            'debugfile': 'shadho_master.debug',
            'password': 'False'
        },
        'backend': {
            'type': 'sql',
            'url': 'sqlite:///:memory:'
        }
    }

    default_config['global']['shadho_dir'] = str(shadho_dir)

    # Load the default config into a ConfigParser. The try block is for
    # Python3+, the except block is for Python 2.7.
    try:
        config = configparser.ConfigParser()
        config.read_dict(default_config)
    except AttributeError:
        config = configparser.RawConfigParser()
        for section in default_config:
            config.add_section(str(section))
            for entry in default_config[section]:
                if type(default_config[section]) is dict:
                    config.set(str(section),
                               str(entry),
                               str(default_config[section][entry]))

    # Make sure the path to the .shadhorc file exists.
    if os.path.basename(prefix) != '.shadhorc':
        os.makedirs(prefix, exist_ok=True)
        prefix = os.path.join(prefix, '.shadhorc')
    else:
        os.makedirs(os.path.dirname(prefix), exist_ok=True)

    # Write the .shadhorc.
    with open(prefix, 'w') as f:
        config.write(f)


def install_shadho_files(prefix, sp_prefix):
    """Setup up additional SHADHO files.

    Parameters
    ----------
    prefix : str
        Install directory of the .shadho directory.
    sp_prefix : str
        Path to the site-packages directory where work_queue was installed.
    """
    maj = sys.version_info.major
    min = sys.version_info.minor
    source = os.path.join(prefix, 'lib', f'python{maj}.{min}', 'site-packages')
    shutil.copy(os.path.join(source, 'work_queue.py'), sp_prefix)
    shutil.copy(os.path.join(source, '_work_queue.so'), sp_prefix)


def main():
    args = parse_args()

    prefix = args.prefix
    if '~' in prefix:
        prefix = os.path.expanduser(prefix)

    if not os.path.isfile(os.path.join(prefix, 'bin', 'work_queue_worker')):
        prefix, sp_prefix = install_workqueue(prefix, user=args.user)
    else:
        maj = sys.version_info.major
        min = sys.version_info.minor

        sp_prefix = os.path.join('lib',
                                 'python{}.{}'.format(maj, min),
                                 'site-packages')
        if args.user:
            sp_prefix = os.path.join(site.USER_BASE, sp_prefix)
        else:
            sp_prefix = os.path.join(sys.prefix, sp_prefix)

    if prefix is not None:
        install_shadho_files(prefix, sp_prefix)

    homedir = str(pathlib.Path().home())
    write_shadhorc(homedir, prefix)


if __name__ == '__main__':
    main()



"""Main driver for the SHADHO framework.

Classes
-------
Shadho
    Driver class for local and distributed hyperparameter optimization.
"""
from .configuration import ShadhoConfig
from .hardware import ComputeClass
from .managers import create_manager

from collections import OrderedDict
import json
import os
import tarfile
import tempfile
import time

import numpy as np
import pyrameter
import scipy.stats

from pyrameter.optimizer import FMin


def shadho():
    pass


class Shadho(pyrameter.FMin):
    """Optimize hyperparameters using specified hardware.

    Parameters
    ----------
    cmd : str or function
        The command to run on remote workers or function to run locally.
    spec : dict
        The specification defining search spaces.
    files : list of str or WQFile
        The files to send to remote workers for task execution.
    use_complexity : bool, optional
        If True, use the complexity heuristic to adjust search proportions.
    use_priority : bool, optional
        If True, use the priority heuristic to adjust search proportions.
    timeout : int, optional
        Number of seconds to search for.
    max_queued_tasks : int, optional
        Number of tasks to queue at a time.
    await_pending : bool, optional
        If True, wait for all running tasks to complete after `timeout`.
        Default: False
    max_resubmissions: int, optional
        Maximum number of times to resubmit a particular parameterization for
        processing if task failure occurs. Default is not to resubmit.

    Attributes
    ----------
    config : `shadho.config.ShadhoConfig`
        Global configurations for shadho.
    backend : `pyrameter.ModelGroup`
        The data view/storage backend. This backend keeps track of all
    manager : `shadho.managers.workqueue.WQManager`
        The distributed task manager to use.
    ccs : list of `shadho.hardware.ComputeClass`
        The types of hardware to expect during optimization. If not supplied,
        tasks are run on the first available worker.
    use_complexity : bool
        If True, use the complexity heuristic to adjust search proportions.
    use_priority : bool
        If True, use the priority heuristic to adjust search proportions.
    timeout : int
        Number of seconds to search for.
    max_queued_tasks : int
        Maximum number of tasks to enqueue at a time.
    await_pending : bool
        If True, wait for all running tasks to complete after `timeout`.
    max_evals : int
        The maximum number of evaluations for a grid point in grid search. Only
        used when the hyperparameter search has no infinite (continuous)
        domains. If ``None``, search until ``timeout``; otherwise, stop early
        once every grid point has been evaluated ``max_evals`` times.
    max_resubmissions: int
        Maximum number of times to resubmit a particular parameterization for
        processing if task failure occurs. Default is not to resubmit.

    Notes
    -----
    To enable configuration after intialization, ``backend`` and ``manager``
    are created when `Shadho.run` is called.

    """

    def __init__(self, exp_key, cmd, spec, method='random', backend='results.json',
                 files=None, use_complexity=True, use_uncertainty=True,
                 timeout=600, max_queued_tasks=100, await_pending=False,
                 max_evals=None, max_resubmissions=0, save_frequency=10,
                 hyperparameters_per_task=1):
        super().__init__(exp_key, spec, method, backend, max_evals=max_evals)
        # self.exp_key = exp_key
        self.config = ShadhoConfig()
        self.cmd = cmd
        if not isinstance(cmd, str):
            self.config.manager = 'local'
        else:
            self.config.workqueue.name = exp_key
        # self.spec = spec
        # self.method = method
        self.use_complexity = use_complexity
        self.use_uncertainty = use_uncertainty
        self.timeout = timeout if timeout is not None and timeout >= 0 \
                       else float('inf')
        self.max_queued_tasks = max_queued_tasks
        self.max_resubmissions = max_resubmissions
        # self.max_evals = max_evals
        self.await_pending = await_pending
        self.save_frequency = save_frequency
        self.hyperparameters_per_task = hyperparameters_per_task \
            if isinstance(hyperparameters_per_task, int) \
            and hyperparameters_per_task > 0 \
            else 1

        self.ccs = OrderedDict()

        self.files = []
        if files is not None:
            for f in files:
                self.files.append(f)
        self.assignments = {}

        self.__tmpdir = tempfile.mkdtemp(prefix='shadho_', suffix='_output')

        self.add_input_file(
            os.path.join(os.path.dirname(__file__), 'worker.py'),
            remotepath=self.config.wrapper)

        self.add_input_file(
            os.path.join(os.path.dirname(__file__), 'utils.py'),
            remotepath=self.config.utils)

        self.config.save_config(self.__tmpdir)
        self.add_input_file(os.path.join(self.__tmpdir, '.shadhorc'))
        # self.backend = backend

    def __del__(self):
        if hasattr(self, '__tmpdir') and self.__tmpdir is not None:
            os.rmdir(self.__tmpdir)

    def add_input_file(self, localpath, remotepath=None, cache=True):
        """Add an input file to the global file list.

        Parameters
        ----------
        localpath : str
            Path to the file on the local filesystem.
        remotepath : str, optional
            Path to write the file to on the remote worker. If omitted, the
            basename of ``localpath`` (e.g. "foo/bar.baz" => "bar.baz").
        cache : bool, optional
            Whether to cache the file on the remote worker. If True (default),
            will be cached on the worker between tasks, reducing network
            transfer overhead. If False, will be re-transferred to the worker
            on each task.
        """
        self.files.append((localpath, remotepath, 'input', cache))

    def add_output_file(self, localpath, remotepath=None, cache=False):
        """Add an input file to the global file list.

        Output files are expected to be discovered on the remote worker after a
        task has completed. They are returned to the `shadho.Shadho` instance
        and will be stored for further review without additional processing.

        Parameters
        ----------
        localpath : str
            Path to the file on the local filesystem.
        remotepath : str, optional
            Path to write the file to on the remote worker. If omitted, the
            basename of ``localpath`` (e.g. "foo/bar.baz" => "bar.baz").
        cache : bool, optional
            Whether to cache the file on the remote worker. It is recommended
            that this be set to False for output files.

        Notes
        -----
        `shadho.Shadho` automatically parses the output file specified in
        ``.shadhorc``, so and output file added through this method will not be
        processed, but rather stored for later review.
        """
        self.files.append((localpath, remotepath, 'output', cache))

    def add_compute_class(self, name, resource, value, max_queued_tasks=100):
        """Add a compute class representing a set of consistent recources.

        Parameters
        ----------
        name : str
            The name of this set of compute resources.
        resource : str
            The resource to match, e.g. gpu_name, cores, etc.
        value
            The value of the resource that should be matched, e.g. "TITAN X
            (Pascal)", 8, etc.
        max_queued_tasks : int, optional
            The maximum number of tasks to queue for this compute class,
            default 100.
        """
        cc = ComputeClass(name, resource, value, 2 * max_queued_tasks, None)
        # cc.optimizer.method = self.method
        self.ccs[cc.id] = cc

    # def load(self):
    #     if not hasattr(self, 'backend') or not isinstance(self.backend, FMin):
    #         self.backend = FMin(self.exp_key, self.spec, self.method,
    #                             self.backend, max_evals=self.max_evals)
    #     self.backend.load()

    # def plot_objective(self, show=True, save=False, filename=None):
    #     try:
    #         self.backend.plot_objective()
    #     except AttributeError:
    #         print('Could not find trials! Have you run a search or loaded results?')

    def run(self):
        """Search hyperparameter values on remote workers.

        Generate and evaluate hyperparameters using the selected task manager
        and search strategy. Hyperparameters will be evaluated until timeout,
        and the optimal set will be printed to screen.

        Notes
        -----
        If `self.await_pending` is True, Shadho will continue to evaluate
        hyperparameters in the queue without generating new hyperparameter
        values. This will continue until the queue is empty and all tasks have
        returned.
        """
        # Set up the task manager as defined in `shadho.managers`
        if not hasattr(self, 'manager'):
            self.manager = create_manager(
                manager_type=self.config.manager,
                config=self.config,
                tmpdir=self.__tmpdir)

        # Set up the backend hyperparameter generation and database
        # if not isinstance(self.backend, FMin):
        #     self.backend = FMin(self.exp_key, self.spec, self.method,
        #                         self.backend, max_evals=self.max_evals)

        # If no ComputeClass was created, create a dummy class.
        if len(self.ccs) == 0:
            cc = ComputeClass('all', None, None, self.max_queued_tasks, super())
            self.ccs[cc.id] = cc
        else:
            for cc in self.ccs.values():
                cc.optimizer = self.copy()

        # Set up intial model/compute class assignments.
        self.assign_to_ccs()

        start = time.time()
        elapsed = 0
        exhausted = False
        try:
            # Run the search until timeout or until all tasks complete
            while elapsed < self.timeout and not exhausted and (elapsed == 0 or not self.manager.empty()):
                # Generate hyperparameters and a flag to continue or stop
                stop = self.generate()
                if not stop:
                    # Run another task and await results
                    result = self.manager.run_task()
                    if result is not None:
                        # If a task returned post-process as a success or fail
                        if len(result) == 3:
                            self.success(*result)  # Store and move on
                        else:
                            self.failure(*result)  # Resubmit if asked
                    # Checkpoint the results to file or DB at some frequency
                    # if self.backend.trial_count % self.save_frequency == 0:
                    if self.trial_count % self.save_frequency == 0:
                        # self.backend.save()
                        self.save()
                    # Update the time for timeout check
                    elapsed = time.time() - start
                    # exhausted = all([ss.done for ss in self.backend.searchspaces])
                    exhausted = all([ss.done for ss in self.searchspaces])
                else:
                    break

            # self.backend.save()
            self.save()

            # If requested, continue the loop until all tasks return
            if self.await_pending:
                while not self.manager.empty():
                    result = self.manager.run_task()
                    if result is not None:
                        if len(result) == 3:
                            self.success(*result)
                        else:
                            self.failure(*result)
                # self.backend.save()
                self.save()

        # On keyboard interrupt, save any results and clean up
        except KeyboardInterrupt:
            if hasattr(self, '__tmpdir') and self.__tmpdir is not None:
                os.rmdir(self.__tmpdir)

        # Save the results and print the optimal set of parameters to  screen
        # self.backend.save()
        self.save()
        # self.backend.summary()
        self.summary()
        # return self.backend.to_dataframes()
        return self.to_dataframes()

    def generate(self):
        """Generate hyperparameter values to test.

        Hyperparameter values are generated from the search space specification
        supplied at instantiation using the requested generation method (i.e.,
        random search, TPE, Gaussian process Bayesian optimization, etc.).

        Returns
        -------
        stop : bool
            If True, no values were generated and the search should stop. This
            facilitates grid-search-like behavior, for example stopping on
            completion of an exhaustive search.

        Notes
        -----
        This method will automatically add a new task to the queue after
        generating hyperparameter values.
        """
        stop = True

        # Generate hyperparameters for every compute class with space in queue
        for cc_id in self.ccs:
            cc = self.ccs[cc_id]
            n = cc.max_queued_tasks - cc.current_tasks

            # Generate enough hyperparameters to fill the queue
            for i in range(n):
                # Get bookkeeping ids and hyperparameter values
                if self.hyperparameters_per_task == 1:
                    trial = cc.generate()
                    trial = super().generate()
                    if trial is not None:
                        # Encode info to map to db in the task tag
                        tag = '.'.join([str(trial.id),
                                        str(trial.searchspace.id),
                                        cc_id])
                        self.manager.add_task(
                            self.cmd,
                            tag,
                            trial.parameter_dict,
                            files=self.files,
                            resource=cc.resource,
                            value=cc.value)

                elif self.hyperparameters_per_task > 1:
                    trial = [cc.generate() for _ in range(self.hyperparameters_per_task)]

                    if not any([t is None for t in trial]):
                        # Encode info to map to db in the task tag
                        tag = '.'.join(['@'.join([str(t.id) for t in trial]),
                                        str(trial[0].searchspace().id),
                                        cc_id])
                        parameters = [t.parameter_dict for t in trial]
                        self.manager.add_task(
                            self.cmd,
                            tag,
                            parameters,
                            files=self.files,
                            resource=cc.resource,
                            value=cc.value)


                # Create a new distributed task if values were generated
                stop = False  # Ensure that the search continues
            cc.current_tasks = cc.max_queued_tasks  # Update to show full queue
        return stop

    def assign_to_ccs(self):
        """Assign trees to compute classes.

        Each independent model in the search (model being one of a disjoint set
        of search domains) is assigned to at least two compute classes based on
        its rank relative to other models. In this way, only a subset of models
        are evaluated on each set of hardware.

        Notes
        -----
        This method accounts for differing counts of models and compute
        classes, adjusting for a greater number of models, a greater number of
        compute classes, or equal counts of models and compute classes.

        See Also
        --------
        `shadho.ComputeClass`
        `pyrameter.ModelGroup`
        """
        # If only one compute class exists, do nothing. If multiple compute
        # classes exist, heuristically assign search trees to CCs. If no
        # compute classes exist, create a dummy to wrap the search.
        if len(self.ccs) > 1:
            # Sort models in the search by complexity, priority, or both and
            # get the updated order.
            # self.backend.sort_spaces(use_complexity=self.use_complexity,
            self.sort_spaces(use_complexity=self.use_complexity,
                             use_uncertainty=self.use_uncertainty)

            # Clear the current assignments
            for cc in self.ccs:
                cc.clear()

            # Determine if the number of compute classes or the number of
            # model ids is larger
            ccids = list(self.ccs.keys())
            # larger = self.backend.searchspaces \
            #          if len(self.backend.searchspaces) >= len(ccids) \
            #          else ccids
            # smaller = ccids if larger == model_ids \
            #           else self.backend.searchspaces
            larger = self.searchspaces \
                     if len(self.searchspaces) >= len(ccids) \
                     else ccids
            smaller = ccids if larger == len(self.searchspaces) \
                      else self.searchspaces

            # Assign models to CCs such that each model is assigned to at
            # least two CCs.

            # Steps between `smaller` index increment
            x = float(len(larger)) / float(len(smaller))
            y = x - 1  # Current step index (offset by 1 for 0-indexing)
            j = 0  # Current index of `smaller`
            m = len(smaller) / 2  # Halfway point for second assignment
            n = len(larger) / 2  # Halfway point for second assignment

            for i in range(len(larger)):
                # If at a step point for `smaller` increment the index
                if i > np.ceil(y):
                    j += 1
                    y += x

                # Add the model to the current CC. If i <= n, add the model to
                # the next CC as well; if i > n, add to the previous CC.
                if smaller[j] in self.ccs:
                    self.ccs[smaller[j]].add_searchspace(larger[i])
                    if j < m:
                        self.ccs[smaller[j + 1]].add_model(
                            self.backend[larger[i]])
                    else:
                        self.ccs[smaller[j - 1]].add_model(
                            self.backend[larger[i]])
                else:
                    self.ccs[larger[i]].add_model(smaller[j])
                    if i < n:
                        self.ccs[larger[i + 1]].add_model(
                            self.backend[smaller[j]])
                    else:
                        self.ccs[larger[i - 1]].add_model(
                            self.backend[smaller[j]])
        elif len(self.ccs) == 0:
            cc = ComputeClass('all', None, None, self.max_queued_tasks, super())
            self.ccs[cc.id] = cc

    def success(self, tag, loss, results):
        """Handle successful task completion.

        Parameters
        ----------
        tag : str
            The task tag, encoding the result id, model id, and compute class
            id as ``<result_id>.<model_id>.<cc_id>``.
        loss : float
            The loss value associated with this result.
        results : dict
            Additional metrics to be included with this result.

        Notes
        -----
        This method will trigger a model/compute class reassignment in the
        event that storing the result caused the model's priority to be
        updated.
        """
        # Get bookkeeping information from the task tag
        trial_id, ss_id, ccid = tag.split('.')

        if not isinstance(results, list):
            results['compute_class'] = {
                'id': ccid,
                'name': self.ccs[ccid].name,
                'value': self.ccs[ccid].value
            }
        else:
            trial_id = trial_id.split('@')
            ccdata = {
                'id': ccid,
                'name': self.ccs[ccid].name,
                'value': self.ccs[ccid].value
            }
            for r in results:
                r['compute_class'] = ccdata


        # Update the DB with the result
        # self.backend.register_result(ss_id, trial_id, loss, results)
        self.register_result(ss_id, trial_id, loss, results)

        # Reassign models to CCs at some frequency
        # n_completed = sum([1 for trial in self.backend.trials.values()
        n_completed = sum([1 for trial in self.trials.values()
                           if trial.status.value == 3])
        if n_completed % 10 == 0:
            self.assign_to_ccs()

        # Update the number of enqueued items
        self.ccs[ccid].current_tasks -= 1

    def failure(self, tag, resub):
        """Handle task failure.

        Parameters
        ----------
        task : `work_queue.Task`
            The failed task to process.

        Notes
        -----
        This method will resubmit failed tasks on request to account for
        potential worker dropout, etc.
        """
        # Get bookkeeping information from the task tag
        trial_id, ss_id, ccid = tag.split('.')

        trials = trial_id.split('@')

        # Determine whether or not to resubmit
        # self.backend.register_result(ss_id, trial_id, objective=None,
        submissions, params = \
            self.register_result(ss_id, trial_id, objective=None,
                                 results=None, errmsg='yes')

        # Resubmit the task if it should be, otherwise update the number of
        # enqueued items.
        if resub and submissions < self.max_resubmissions:
            cc = self.ccs[ccid]
            self.manager.add_task(self.cmd,
                                  tag,
                                  params,
                                  files=self.files,
                                  resource=cc.resource,
                                  value=cc.value)
        else:
            self.ccs[ccid].current_tasks -= 1


# class PopulationShadho(Shadho):
#     def __init__(self, exp_key, cmd, spec, generations=50, method='random',
#                  backend='results.json', files=None, use_complexity=True,
#                  use_uncertainty=True, timeout=600, max_queued_tasks=100,
#                  await_pending=False, max_evals=None, max_resubmissions=0,
#                  save_frequency=10, hyperparameters_per_task=1):
#         super().__init__(
#             exp_key,
#             cmd,
#             spec,
#             method='random',
#             backend='results.json',
#             files=None,
#             use_complexity=True,
#             use_uncertainty=True,
#             timeout=600,
#             max_queued_tasks=100,
#             await_pending=False,
#             max_evals=None,
#             max_resubmissions=0,
#             save_frequency=10,
#             hyperparameters_per_task=1)

#         self.generations = generations

#     def run(self):
#         # Set up the task manager as defined in `shadho.managers`
#         if not hasattr(self, 'manager'):
#             self.manager = create_manager(
#                 manager_type=self.config.manager,
#                 config=self.config,
#                 tmpdir=self.__tmpdir)

#         # Set up the backend hyperparameter generation and database
#         # if not isinstance(self.backend, FMin):
#         #     self.backend = FMin(self.exp_key, self.spec, self.method,
#         #                         self.backend, max_evals=self.max_evals)

#         # If no ComputeClass was created, create a dummy class.
#         if len(self.ccs) == 0:
#             cc = ComputeClass('all', None, None, self.max_queued_tasks, self)
#             self.ccs[cc.id] = cc
#         else:
#             for cc in self.ccs.values():
#                 cc.optimizer = self.backend.copy()

#         # Set up intial model/compute class assignments.
#         self.assign_to_ccs()

#         start = time.time()
#         elapsed = 0
#         generation = 0
#         exhausted = False
#         try:
#             # Run the search until timeout or until all tasks complete
#             while elapsed < self.timeout and generation < self.generations:
#                 # Generate hyperparameters and a flag to continue or stop
#                 stop = self.generate()
#                 if not stop:
#                     # Run another task and await results
#                     pop_size = len(stop)
#                     result = self.manager.run_task()
#                     if result is not None:
#                         # If a task returned post-process as a success or fail
#                         if len(result) == 3:
#                             self.success(*result)  # Store and move on
#                         else:
#                             self.failure(*result)  # Resubmit if asked
#                     # Checkpoint the results to file or DB at some frequency
#                     # if self.backend.trial_count % self.save_frequency == 0:
#                     if self.trial_count % self.save_frequency == 0:
#                         # self.backend.save()
#                         self.save()
#                     # Update the time for timeout check
#                     elapsed = time.time() - start
#                     # exhausted = all([ss.done for ss in self.backend.searchspaces])
#                     exhausted = all([ss.done for ss in self.searchspaces])
#                 else:
#                     break

#             # self.backend.save()
#             self.save()

#             # If requested, continue the loop until all tasks return
#             if self.await_pending:
#                 while not self.manager.empty():
#                     result = self.manager.run_task()
#                     if result is not None:
#                         if len(result) == 3:
#                             self.success(*result)
#                         else:
#                             self.failure(*result)
#                 # self.backend.save()
#                 self.save()

#         # On keyboard interrupt, save any results and clean up
#         except KeyboardInterrupt:
#             if hasattr(self, '__tmpdir') and self.__tmpdir is not None:
#                 os.rmdir(self.__tmpdir)

#         # Save the results and print the optimal set of parameters to  screen
#         # self.backend.save()
#         self.save()
#         # self.backend.summary()
#         self.summary()
#         # return self.backend.to_dataframes()
#         return self.to_dataframes()



"""Utilities for submitting and starting Work Queue workers.


"""
import argparse
import os
import re
import subprocess

from shadho.configuration import ShadhoConfig


def parse_args(args=None):
    p = argparse.ArgumentParser(
        description='Start a Work Queue worker and connect to SHADHO.')

    p.add_argument('-M', '--master', type=str, default='',
        help='name of the Work Queue master to connect to')
    p.add_argument('-u', '--user', type=str, default=os.environ['USER'],
        help='name of the user running the Work Queue master')
    p.add_argument('--timeout', type=int,
        help='amount of time worker idles before exiting')
    p.add_argument('--cores', type=int, default=1,
        help='the number of cores for the worker to use;' +
             ' pass 0 to use all available cores')
    p.add_argument('--feature', type=str, nargs='*', default=[],
        help='user specified feature to advertise, e.g. GPU model name')

    return p.parse_args(args)


def shadho_wq_worker(args=None, config=None):
    """Start a Work Queue worker."""

    if config is None:
        config = ShadhoConfig()

    if args is None:
        cmd_args = ''
    else:
        cmd_args = f'{"-M" if args.master else ""} {args.master} --cores {args.cores}'
        for feature in args.feature:
            cmd_args += f' --feature {feature}'

    if not re.search(r'(^|[\s])-M([\s]|$)', cmd_args):
        cmd_args = ' '.join([cmd_args, '-M', config.workqueue.name]).strip()

    if not re.search(r'[\s]*-M[\s][\S]*' + args.user + r'.*[\s]*', cmd_args):
        print('Replacing')
        cmd_args = re.sub(r'(^|[\s]*)(.*-M[\s])([\S]+)([\s]*.*$)',
                          r'\1\2\3-' + args.user + r'\4',
                          cmd_args)

    executable = os.path.join(config.shadho_dir, 'bin', 'work_queue_worker')

    print(cmd_args)
    subprocess.run([executable] + cmd_args.split(), stderr=subprocess.STDOUT)


def main():
    args = parse_args()
    shadho_wq_worker(args=args)


if __name__ == '__main__':
    main()



def create_manager(manager_type='local', config=None, tmpdir=None):
    print(config.config)
    if manager_type == 'workqueue':
        from .workqueue import WQManager
        return WQManager(
            config.param_file,
            config.output,
            config.result_file,
            config.optimize,
            tmpdir,
            name=config.workqueue.name,
            port=config.workqueue.port,
            shutdown=config.workqueue.shutdown,
            logfile=config.workqueue.logfile,
            debugfile=config.workqueue.debugfile
        )
    else:
        from .local import LocalManager
        return LocalManager(config.optimize)



"""Top-level imports for SHADHO.

At the top level, the SHADHO package should grant access to the `Shadho` object
that drives hyperparameter optimization and the configuration.

See Also
--------
shadho.shadho
    Main driver for hyperparameter optimization.
shadho.configuration
    Utility for configuring SHADHO.

"""

from .shadho import Shadho

__all__ = ['Shadho', 'managers', 'spaces']



"""Work Queue wrapper for hardware-aware distributed task management.

Classes
-------
WQManager
    Work Queue master with additional task packaging logic.
WQFile
WQBuffer
    Work-Queue-conformant representations of files and buffers for task I/O.
"""
import json
import os
import sys
import tarfile

import work_queue as WORKQUEUE

from shadho.utils import ShadhoEncoder, ShadhoDecoder


class WQManager(WORKQUEUE.WorkQueue):
    """Work Queue master with utilities to generate conformant tasks.

    Parameters
    ----------
    param_file : str
        The name of the hyperparameter value file sent to the worker.
    out_file : str
        The name of the expected output file returned from the worker.
    results_file : str
        The name of the expected results file returned from the worker.
    opt_value : str
        The name of the value to optimize on in `results_file`.
    tmpdir : str
        Temporary directory used by SHADHO.
    name : str, optional
        The name of this Work Queue master, 'shadho' by default.
    port : int, optional
        The port that the master will listen on. 9123 by default.
    exclusive : bool, optional
        Whether to allow multiple Work Queue masters in the same process. True
        by default.
    shutdown : bool, optional
        Whether to shut down connected workers when this manager shuts down.
        True by default.
    logfile : str, optional
        Where to write Work Queue logs. ./wq_shadho.log by default.
    debugfile : str, optional
            Where to write Work Queue debug logs. ./wq_shadho.debug by default.

    Attributes
    ----------

    """

    def __init__(self, param_file, out_file, results_file, opt_value, tmpdir,
                 name='shadho', port=9123, shutdown=True,
                 logfile='shadho_wq.log', debugfile='shadho_wq.debug'):
        WORKQUEUE.cctools_debug_flags_set("all")
        WORKQUEUE.cctools_debug_config_file(debugfile)
        WORKQUEUE.cctools_debug_config_file_size(0)

        if os.environ['USER'] not in name:
            name += '-{}'.format(os.environ['USER'])

        super(WQManager, self).__init__(name=name,
                                        port=int(port),
                                        shutdown=shutdown)

        self.specify_log(logfile)

        self.param_file = param_file
        self.out_file = out_file
        self.results_file = results_file
        self.opt_value = opt_value
        self.tmpdir = tmpdir
        self.tasks_submitted = self.stats.tasks_submitted

    def add_task(self, cmd, tag, params, files=None, resource=None, value=None):
        """Create a task for this manager.

        Parameters
        ----------
        cmd : str
            The command to run on the remote worker, e.g. ``echo hello`` or
            ``python script.py``.
        tag : str
            The tag to give this task.
        files : list of `WQFile` or `WQBuffer`, optional
            The input and output files and data buffers that this task will
            send/receive.

        Notes
        -----
        Any command may be passed to the task to run, and the task makes stdout
        and stderr of the command available upon return, regardless of failure.

        See Also
        --------
        `shadho.managers.workqueue.WQFile`
        `shadho.managers.workqueue.WQBuffer`
        `work_queue.Task`
        """
        # Set up the task to run the specified command with its tag as the
        # trailing argument.
        task = WORKQUEUE.Task(' '.join([cmd, tag]))
        task.specify_tag(tag)

        # Set up the input and output file structure of the task. This
        # structure is preserved on the worker.
        if files is None:
            files = []

        for f in files:
            if isinstance(f, tuple):
                f = WQFile(f[0], remotepath=f[1], ftype=f[2], cache=f[3])
            f.add_to_task(task)

        # Set up the output file
        out = WQFile(os.path.join(self.tmpdir,
                                  '.'.join([tag, self.out_file])),
                     remotepath=self.out_file,
                     ftype='output',
                     cache=False)

        # Send the hyperparameters as a JSON string
        buff = WQBuffer(str(json.dumps(params, cls=ShadhoEncoder)),
                        self.param_file,
                        cache=False)

        out.add_to_task(task)
        buff.add_to_task(task)

        # Add information about the expected hardware resource
        if resource is not None:
            if resource == 'cores':
                task.specify_cores(value)
            elif resource == 'feature':
                task.specify_requirement(value)
            else:
                task.specify_resource(resource, value)

        # Submit the task
        self.submit(task)
        self.tasks_submitted = self.stats.tasks_submitted

    def run_task(self):
        """Await the return of a running task.

        Returns
        -------
        tag : str
            The task tag.
        loss : float
            The loss value. Only returned on success.
        results : dict
            Additional results returned by the objective function.
            Only returned on success.
        """
        task = None
        while task is None:
            # Wait for a task to return for 10s
            task = self.wait(timeout=10)

            # Handle task success or failure (no return if `task` is None)
            if task is not None:
                if self.task_succeeded(task):
                    return self.success(task)
                else:
                    return self.failure(task)

    def task_succeeded(self, task):
        """Determine whether or not a task succeeded.

        Parameters
        ----------
        task : work_queue.Task
            The task to check.

        Returns
        -------
        True if the task succeeded, False otherwise.

        Notes
        -----
        The task succeeded if it was not None and Work Queue reported a
        successful return.
        """
        return task is not None and \
            task.result == WORKQUEUE.WORK_QUEUE_RESULT_SUCCESS

    def success(self, task):
        """Handle Work Queue task success.

        Parameters
        ----------
        task : work_queue.Task
            The successful task with results to process.

        Returns
        -------
        result_id : str
            The id of the database entry representing this result.
        cc_id : str
            The id of the compute class that ran this result.
        loss : float
            The value being optimized.
        results : dict
            Other results returned by the task.
        """
        try:
            # Open the result tarfile and get the results file.
            outfile = '.'.join([task.tag, self.out_file])
            result = tarfile.open(os.path.join(self.tmpdir, outfile), 'r')
            resultstr = result.extractfile(self.results_file).read()
            result.close()
        except IOError:
            print("Error opening task {} result".format(rid))

        result = json.loads(resultstr.decode('utf-8'), cls=ShadhoDecoder)
        if isinstance(result, list):
            loss = []
            for r in result:
                loss.append(r[self.opt_value])
                r['submit_time'] = task.submit_time
                r['start_time'] = task.execute_cmd_start
                r['finish_time'] = task.finish_time
        else:
            loss = result[self.opt_value]
            result['submit_time'] = task.submit_time
            result['start_time'] = task.execute_cmd_start
            result['finish_time'] = task.finish_time
        return (task.tag, loss, result)

    def failure(self, task):
        """Handle Work Queue task failure.

        Parameters
        ----------
        task : work_queue.Task
            The failed task to process.

        Returns
        -------
        result_id : str
            The id of the database entry representing this result.
        cc_id : str
            The id of the compute class that ran this result.
        """
        print('Task {} failed with result {} and WQ status {}'
              .format(task.tag, task.result, task.return_status))
        print(task.output)

        resub = (int(task.return_status) == 137)

        #rid, ccid = str(task.tag).split('.')
        return (task.tag, resub)


class WQFile(object):
    """File to send to or receive from a remote worker.

    Parameters
    ----------
    localpath : str
        The path to the file on the local system.
    remotepath : str, optional
        Where to place the file on the remote worker relative to cwd. Set to
        the basename of `localpath` by default (e.g. foo/bar.baz -> bar.baz).
    ftype : {'input', 'output'}
        Whether to send this file to or expect to receive it from the remote
        worker.
    cache : bool, optional
        Whether to cache this file on the remote worker.

    Attributes
    ----------
    localpath : str
        The path to the file on the local system.
    remotepath : str
        Where to place the file on the remote worker relative to remote cwd.
    ftype : {'input', 'output'}
        Whether to send this file to or expect to receive it from the remote
        worker.
    cache : bool
        Whether to cache this file on the remote worker.

    """

    TYPES = {
        'input': WORKQUEUE.WORK_QUEUE_INPUT,
        'output': WORKQUEUE.WORK_QUEUE_OUTPUT,
    }

    CACHE = {
        True: WORKQUEUE.WORK_QUEUE_CACHE,
        False: WORKQUEUE.WORK_QUEUE_NOCACHE
    }

    def __init__(self, localpath, remotepath=None, ftype='input', cache=True):
        if ftype != 'output' and not os.path.isfile(localpath):
            raise IOError("{} does not exist.".format(localpath))
        else:
            self.localpath = localpath

        self.remotepath = remotepath if remotepath is not None \
                          else os.path.basename(localpath)

        self.ftype = WQFile.TYPES[ftype]
        self.cache = WQFile.CACHE[cache]

    def add_to_task(self, task):
        """Add the file to a task.

        Parameters
        ----------
        task : `work_queue.Task`
            The task to add the file to.
        tag : str, optional
            The task tag.

        Notes
        -----
        If this file is an output file, `tag` is prepended to the local path
        name, then
        """
        task.specify_file(self.localpath,
                          remote_name=self.remotepath,
                          type=self.ftype,
                          cache=self.cache)


class WQBuffer(object):
    """File to send to or receive from a remote worker.

    Parameters
    ----------
    buffer : str
        The string buffer to send to the remote task.
    remotepath : str, optional
        Where to place the file on the remote worker relative to cwd. Set to
        the basename of `localpath` by default (e.g. foo/bar.baz -> bar.baz).
    cache : bool, optional
        Whether to cache this file on the remote worker.

    Attributes
    ----------
    buffer : str
        The string buffer to send to the remote task.
    remotepath : str
        Where to place the file on the remote worker relative to remote cwd.
    cache : bool
        Whether to cache this file on the remote worker.

    Notes
    -----
    Buffers are most useful for task-specific data (e.g. hyperparameter values)
    and as such are always treated as non-caching input files.

    """

    def __init__(self, buffer, remotepath, cache=False):
        self.buffer = str(buffer)
        self.remotepath = str(remotepath)
        self.cache = WQFile.CACHE[cache]
        self.ftype = 'buffer'

    def add_to_task(self, task):
        """Add the buffer to a task.

        Parameters
        ----------
        task : `work_queue.Task`
            The task to add the file to.
        tag : str, optional
            The task tag.

        Notes
        -----
        `tag` is currently unused, included as a placeholder to match the
        signature of `shadho.managers.workqueue.WQFile.add_to_task`.
        """
        task.specify_buffer(self.buffer, self.remotepath, self.cache)



"""Hyperparameter search space domain definitions.

Functions
---------
scope
    Create a scope to join multiple search domains.
uniform
    Create a continuous uniform distribution.
ln_uniform
    Create a natural-log-scaled continuous uniform distribution.
log10_uniform
    Create a log-10-scaled continuous uniform distribution.
log2_uniform
    Create a log-2-scaled continuous uniform distribution.
normal
    Create a normal distribution.
ln_normal
    Create a natural-log-scaled normal distribution.
log10_normal
    Create a log-10-scaled normal distribution.
log2_normal
    Create a log-2-scaled normal distribution.
randint
    Create a discrete integer domain.
log10_randint
    Create a log-10-scaled discrete integer domain.
log2_randint
    Create a log-2-scaled discrete integer domain.
choice
    Create a domain over a set of arbitrary categorical values.

See Also
--------
shadho.scaling
pyrameter.domain
"""
from shadho.scaling import linear, ln, log_10, log_2

import numpy as np
from pyrameter.domains import *
from pyrameter.specification import Specification
import scipy.stats


def scope(exclusive=False, optional=False, **kwargs):
    """Create a scope for hierarchical domain definitions.

    Hyperparameter search spaces can be set up as trees, with related
    hyperparameters grouped in subtrees. Scopes enable this type of
    organization.

    Parameters
    ----------
    exclusive : bool, optional
        If True, only generate a single hyperparameter (or sub-scope) from this
        scope at a time. Default ``False``.
    optional : bool, optional
        If True, either generate a set of hyperparameters from this scope or
        ignore the scope entirely. Default ``False``.
    domains
        Key/value pairs of domain names and valid SHADHO domains in this scope.

    Returns
    -------
    scope : `pyrameter.specification.Specification`
        The scope with all sub-domains.
    """
    return Specification(**kwargs)


# Uniform distribution
def uniform(lo, hi, **kwargs):
    """Continuous uniform distribution.

    The distribution is defined in range [``lo``, ``hi``].

    Parameters
    ----------
    lo : int, float
        Lower bound of the distribution.
    hi : int, float
        Upper bound of the distribution.

    Returns
    -------
    domain : `pyrameter.ContinuousDomain`
    """
    return ContinuousDomain(scipy.stats.uniform, loc=lo, scale=np.abs(hi - lo),
                            callback=linear, **kwargs)


def ln_uniform(lo, hi, **kwargs):
    """Natural log-scaled continuous uniform distribution.

    The distribution is defined in range [``lo``, ``hi``]. Sampled values ``x``
    are scaled by ``e**x`` before being returned.

    Parameters
    ----------
    lo : int, float
        Lower bound of the distribution.
    hi : int, float
        Upper bound of the distribution.

    Returns
    -------
    domain : `pyrameter.ContinuousDomain`
    """
    return ContinuousDomain(scipy.stats.uniform, loc=lo, scale=np.abs(hi - lo), callback=ln,
                            **kwargs)


def log10_uniform(lo, hi, **kwargs):
    """Log 10-scaled continuous uniform distribution.

    The distribution is defined in range [``lo``, ``hi``]. Sampled values ``x``
    are scaled by ``10**x`` before being returned.

    Parameters
    ----------
    lo : int, float
        Lower bound of the distribution.
    hi : int, float
        Upper bound of the distribution.

    Returns
    -------
    domain : `pyrameter.ContinuousDomain`
    """
    return ContinuousDomain(scipy.stats.uniform, loc=lo, scale=np.abs(hi - lo),
                            callback=log_10, **kwargs)


def log2_uniform(lo, hi, **kwargs):
    """Log 2-scaled continuous uniform distribution.

    The distribution is defined in range [``lo``, ``hi``]. Sampled values ``x``
    are scaled by ``2**x`` before being returned.

    Parameters
    ----------
    lo : int, float
        Lower bound of the distribution.
    hi : int, float
        Upper bound of the distribution.

    Returns
    -------
    domain : `pyrameter.ContinuousDomain`
    """
    return ContinuousDomain(scipy.stats.uniform, loc=lo, scale=np.abs(hi - lo),
                            callback=log_2, **kwargs)


# Normal distribution
def normal(mu, sigma, **kwargs):
    """Continuous normal distribution.

    Gaussian distribution parameterized with expected value ``mu`` and standard
    deviation ``sigma``.

    Parameters
    ----------
    mu : int, float
        Expected value of the distribution.
    sigma : int, float
        Standard deviation of the distribution.

    Returns
    -------
    domain : `pyrameter.ContinuousDomain`
    """
    return ContinuousDomain(scipy.stats.norm, loc=lo, scale=hi,
                            callback=linear, **kwargs)


def ln_normal(mu, sigma, **kwargs):
    """Natural log-scaled continuous normal distribution.

    Gaussian distribution parameterized with expected value ``mu`` and standard
    deviation ``sigma``. Sampled values ``x`` are scaled by ``e**x`` before
    being returned.

    Parameters
    ----------
    mu : int, float
        Expected value of the distribution.
    sigma : int, float
        Standard deviation of the distribution.

    Returns
    -------
    domain : `pyrameter.ContinuousDomain`
    """
    return ContinuousDomain(scipy.stats.norm, loc=lo, scale=hi, callback=ln,
                            **kwargs)


def log10_normal(mu, sigma, **kwargs):
    """Log 10-scaled continuous normal distribution.

    Gaussian distribution parameterized with expected value ``mu`` and standard
    deviation ``sigma``. Sampled values ``x`` are scaled by ``10**x`` before
    being returned.

    Parameters
    ----------
    mu : int, float
        Expected value of the distribution.
    sigma : int, float
        Standard deviation of the distribution.

    Returns
    -------
    domain : `pyrameter.ContinuousDomain`
    """
    return ContinuousDomain(scipy.stats.norm, loc=lo, scale=hi,
                            callback=log_10, **kwargs)


def log2_normal(mu, sigma, **kwargs):
    """Log 2-scaled continuous normal distribution.

    Gaussian distribution parameterized with expected value ``mu`` and standard
    deviation ``sigma``. Sampled values ``x`` are scaled by ``2**x`` before
    being returned.

    Parameters
    ----------
    mu : int, float
        Expected value of the distribution.
    sigma : int, float
        Standard deviation of the distribution.

    Returns
    -------
    domain : `pyrameter.ContinuousDomain`
    """
    return ContinuousDomain(scipy.stats.norm, loc=lo, scale=hi,
                            callback=log_2, **kwargs)


# Randint distributions
def randint(lo, hi, step=1):
    """Uniform distribution over a sequence of integers.

    The distribution is over the integer sequence [``lo``, ``hi``) with step
    ``step`` spacing.

    Parameters
    ----------
    lo : int
        Lower bound of the sequence.
    hi : int
        Upper bound of the sequence.
    step : int, optional
        Step between members of the sequence.

    Returns
    -------
    domain : `pyrameter.DiscreteDomain`
    """
    return DiscreteDomain(list(range(lo, hi, step)))


def ln_randint(lo, hi, step=1):
    """Natural log-scaled uniform distribution over a sequence of integers.

    The distribution is over the integer sequence [``lo``, ``hi``) with step
    ``step`` spacing. Sampled values ``x`` are scaled by ``e**x`` before
    being returned.

    Parameters
    ----------
    lo : int
        Lower bound of the sequence.
    hi : int
        Upper bound of the sequence.
    step : int, optional
        Step between members of the sequence.

    Returns
    -------
    domain : `pyrameter.DiscreteDomain`
    """
    return DiscreteDomain([ln(i) for i in range(lo, hi, step)])


def log10_randint(lo, hi, step=1):
    """Log 10-scaled uniform distribution over a sequence of integers.

    The distribution is over the integer sequence [``lo``, ``hi``) with step
    ``step`` spacing. Sampled values ``x`` are scaled by ``10**x`` before
    being returned.

    Parameters
    ----------
    lo : int
        Lower bound of the sequence.
    hi : int
        Upper bound of the sequence.
    step : int, optional
        Step between members of the sequence.

    Returns
    -------
    domain : `pyrameter.DiscreteDomain`
    """
    return DiscreteDomain([log_10(i) for i in range(lo, hi, step)])


def log2_randint(lo, hi, step=1):
    """Log 2-scaled uniform distribution over a sequence of integers.

    The distribution is over the integer sequence [``lo``, ``hi``) with step
    ``step`` spacing. Sampled values ``x`` are scaled by ``2**x`` before
    being returned.

    Parameters
    ----------
    lo : int
        Lower bound of the sequence.
    hi : int
        Upper bound of the sequence.
    step : int, optional
        Step between members of the sequence.

    Returns
    -------
    domain : `pyrameter.DiscreteDomain`
    """
    return DiscreteDomain([log_2(i) for i in range(lo, hi, step)])


# Choice
def choice(choices):
    """Random selection from an arbitrary set of values.

    This domain performs randint sampling from a set of aarbitrary values,
    allowing for randomly selecting strings, tuples, arbitrary collections of
    numbers, etc.

    Parameters
    ----------
    choices : list
        List of values to choose from.

    Returns
    -------
    domain : `pyrameter.DiscreteDomain`
    """
    return DiscreteDomain(choices)


# Exhaustive (a.k.a. Grid Search)
def exhaustive(choices):
    """Exhaustive search over a set of categorical values.

    Parameters
    ----------
    choices : list
        List of values to choose from.

    Returns
    -------
    domain : pyrameter.ExhaustiveDomain
    """
    return ExhaustiveDomain(choices)


def repeat(domain, repetitions, split=True):
    """Repeat a given domain a number of times.

    It can be helpful to repeat a domain a number of times, for example when
    iterating over a sequence of search spaces defined over the same values.
    If desired, the resulting domain can be split into ``repetitions``
    separate domains of length [1, 2, ..., ``repetitions``].

    Parameters
    ----------
    domain
        The domain to repeat, any valid SHADHO domain.
    repetitions : int
        The number of times to repeat the domain.
    split : bool
        If True, split the domain into ``repetitions`` sequences of domains of
        length [1, 2, ..., ``repetitions``]. If False, behaves like a single
        sequence. Default: True.

    Returns
    -------
    domain : pyrameter.RepeatedDomain
    """
    return RepeatedDomain(domain, repetitions, split=split)


def sequential(domains):
    """Construct a domain that behaves like a sequence when sampled.

    This domain consists of a tuple of domains that are sampled independently.
    When sampled, this will return a sequence of values in the same order as
    the provided domain sequence.

    Parameters
    ----------
    domains
        A list, tuple, or other sequence of valid SHADHO domains.
    
    Returns
    -------
    domain : pyrameter.SequenceDomain
    """
    return SequenceDomain(domains)


def dependent(domain, callback=None):
    return DependentDomain(domain, callback=lambda x: x if callback is None else callback)



"""Task management for local, serial hyperparameter optimization.

Classes
-------
LocalManager
    Task manager for running tasks on the local machine.
LocalTask
    Wrapper for running a task locally.
"""
from collections import deque
import json
import time
import uuid


class TaskFailureError(Exception):
    """Raised when a task fails for any reason."""
    def __init__(self, e):
        super(TaskFailureError, self).__init__(str(e))


class LocalManager(object):
    """Task manager for running tasks on the local machine.

    Parameters
    ----------
    opt_value : str
        The value to search for in a task result.

    Attributes
    ----------
    tasks : list of LocalTask
        The tasks to run locally.
    opt_value : str
        The value to search for in a task result.
    """
    def __init__(self, opt_value):
        self.tasks = deque()
        self.opt_value = opt_value
        self.tasks_submitted = 0

    def empty(self):
        """Determine if the task queue is empty.

        Returns
        -------
        empty : bool
            True if the queue contains no tasks, False otherwise.
        """
        return len(self.tasks) == 0

    def add_task(self, cmd, tag, params, files=None, resource=None,
                 value=None):
        """Add a task to the task list.

        Parameters
        ----------
        cmd : callable
            The objective function.
        tag : str
            The tag to give this task.
        params : dict
            The hyperparameter value to supply to the task.
        files : optional
            Placeholder to match other managers.
        resource : optional
            Placeholder to match other managers.
        value : optional
            Placeholder to match other managers.
        """
        # Create the LocalTask instance
        task = LocalTask(cmd, tag, params)

        # Enqueue the new task
        self.tasks.append(task)
        self.tasks_submitted += 1

    def run_task(self):
        """Run the next task on the task list and return its result.

        This function chooses the next task to run from the task list and
        attempts to run it. Any encountered errors are repoted to the user.

        Returns
        -------
        result_id : str
            The id of the database entry representing this result.
        cc_id : str
            The id of the compute class that ran this result.
        loss : float
            Only returned on success. The value being optimized.
        results : dict
            Only returned on success. Other results returned by the task.
        """
        # Attempt to get the next task
        try:
            task = self.tasks.popleft()
        except IndexError:
            task = None

        # If there is a task, try to run it
        if task is not None:
            result = None

            # Set up the task tag for return
            ret = [task.tag]

            # Try to run the task, and attempt to catch and report any errors
            try:
                result = task.run()

                # Package the result to return
                if result is not None:
                    # Handle the case of multiple values and single value being returned
                    if not isinstance(result, list):
                        if isinstance(result, dict):
                            ret.append(result[self.opt_value])
                            ret.append(result)
                        elif isinstance(result, float):
                            ret.append(result)
                            ret.append({self.opt_value: result})
                        else:
                            raise ValueError
                    else:
                        optima = []
                        results = []

                        for r in result:
                            if isinstance(r, dict):
                                optima.append(r[self.opt_value])
                                results.append(r)
                            elif isinstance(r, float):
                                optima.append(r)
                                results.append({self.opt_value: r})
                            else:
                                raise ValueError

                        ret.append(optima)
                        ret.append(results)
            except TaskFailureError as e:
                print("Error: Task failed due to the following error:")
                print(str(e))
            except ValueError:
                print("Error: Invalid task result {}".format(result))
                print("Task results must be of type float")
                print("or dict with a float in key {}".format(self.opt_value))
            except KeyError:
                print("Error: Result {} does not contain the value {} to optimize."
                      .format(result, self.opt_value))
                print("Please check your function to ensure that it returns the")
                print("correct value.")

            return tuple(ret) if result is not None else None


class LocalTask(object):
    """A task to run using supplied hyperparameters.

    Parameters
    ----------
    cmd : function or callable object
        The command to run.
    tag : str
        Tag with metadata about the task.
    params : dict
        The parameters to supply to the task.

    Attributes
    ----------
    id : str
        Internal task id.
    cmd : function or callable object
        The command to run.
    tag : str
        Tag with metadata about the task.
    params : dict
        The parameters to supply to the task.
    """
    def __init__(self, cmd, tag, params):
        self.id = str(uuid.uuid4())
        self.cmd = cmd
        self.tag = tag
        self.params = params

    def run(self):
        """Run the task and return the result.

        Returns
        -------
        result : float or dict
            The result of the task, either a float representing the loss or a
            dictionary containing the loss and (possibly) other results.
        """
        try:
            result = self.cmd(self.params)
        except Exception as e:
            raise TaskFailureError(str(e))
        return result



import json

import numpy as np


class ShadhoEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return {
                '__data': list(obj.ravel()),
                '__dtype': str(obj.dtype),
                '__shape': obj.shape
            }
        else:
            return super(ShadhoEncoder, self).default(obj)


class ShadhoDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super(ShadhoDecoder, self).__init__(
            *args, object_hook=self.object_hook, **kwargs)

    def object_hook(self, obj):
        if isinstance(obj, dict) and sorted(obj.keys()) == ['__data', '__dtype', '__shape']:
            arr = np.array(obj['__data']).astype(obj['__dtype']).reshape(obj['__shape'])
            return arr
        else:
            return obj



#!/usr/bin/env python
import os
from setuptools import setup
from setuptools.command.install import install

LONG_DESCRIPTION = ""
with open(os.path.join(os.path.dirname(__file__), 'README.md'), 'r') as f:
    LONG_DESCRIPTION = f.read()


setup(
    name='shadho',
    version='0.3.0.post2',
    description='Hyperparameter optimizer with distributed hardware at heart',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/jeffkinnison/shadho',
    author='Jeff Kinnison',
    author_email='jkinniso@nd.edu',
    python_requires='>=3.5',
    packages=['shadho',
              'shadho.installers',
              'shadho.managers',
              'shadho.workers',],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: POSIX',
        'Operating System :: Unix',
    ],
    keywords='machine_learning hyperparameters distributed_computing',
    install_requires=[
        'numpy',
        'scipy',
        'scikit-learn',
        'pyrameter'
    ],
    tests_require=['pytest'],
    include_package_data=True,
)
